"""Eeg class to manipulate EEG data and associated metadata. The class interfaces with EDF files."""

import datetime
import enum
import os
import re
from typing import TypedDict

import numpy as np
import pandas as pd
import pyedflib
import resampy


class FileFormat(str, enum.Enum):
    CSV = "csv"
    CSV_GZIP = "csv.gzip"
    EDF = "edf"
    PARQUET_GZIP = "parquet.gzip"


class Eeg:
    class Montage(str, enum.Enum):
        UNIPOLAR = "unipolar"
        BIPOLAR = "bipolar"

    BIPOLAR_DBANANA = (
        "Fp1-F3",
        "F3-C3",
        "C3-P3",
        "P3-O1",
        "Fp1-F7",
        "F7-T3",
        "T3-T5",
        "T5-O1",
        "Fz-Cz",
        "Cz-Pz",
        "Fp2-F4",
        "F4-C4",
        "C4-P4",
        "P4-O2",
        "Fp2-F8",
        "F8-T4",
        "T4-T6",
        "T6-O2",
    )
    ELECTRODES_10_20 = (
        "Fp1",
        "F3",
        "C3",
        "P3",
        "O1",
        "F7",
        "T3",
        "T5",
        "Fz",
        "Cz",
        "Pz",
        "Fp2",
        "F4",
        "C4",
        "P4",
        "O2",
        "F8",
        "T4",
        "T6",
    )

    class SignalHeader(TypedDict):
        label: str  # channel label (string, <= 16 characters, must be unique)
        dimension: str  # physical dimension (e.g., mV) (string, <= 8 characters)
        sample_frequency: int  # number of samples per second
        physical_max: float  # maximum physical value
        physical_min: float  # minimum physical value
        digital_max: int  # maximum digital value (-2**15 <= x < 2**15)
        digital_min: int  # minimum digital value (-2**15 <= x < 2**15)

    DEFAULT_SIGNAL_HEADER: SignalHeader = {
        "label": "channel",
        "dimension": "uV",
        "sample_frequency": 256,
        "physical_max": 1e6,
        "physical_min": -1e6,
        "digital_max": 2**14,
        "digital_min": -(2**14),
    }

    class FileHeader(TypedDict):
        technician: str
        recording_additional: str
        patient_name: str
        patient_additional: str
        patient_code: str
        equipment: str
        admincode: str
        gender: int  # 1 is male, 0 is female
        recording_start_time: datetime.datetime
        birthdate: datetime.date

    DEFAULT_FILE_HEADER: FileHeader = {
        "technician": "",
        "recording_additional": "",
        "patient_name": "",
        "patient_additional": "",
        "patient_code": "",
        "equipment": "",
        "admincode": "",
        "gender": 0,
        "recording_start_time": datetime.datetime(1970, 1, 1),
        "birthdate": datetime.date(1970, 1, 1),
    }

    def __init__(
        self,
        data,
        channels: tuple[str],
        fs: int,
        montage: Montage = Montage.UNIPOLAR,
        signalHeader: SignalHeader = DEFAULT_SIGNAL_HEADER,
        fileHeader: FileHeader = DEFAULT_FILE_HEADER,
    ):
        """Initiate an EEG instance

        Args:
            data (NDArray[Shape['*, *'], float]): data array, rows are channels, columns are samples in time
            channels (tuple[str]): tuple of channels as strings.
            fs (int): Sampling frequency.
            montage (MontageType, optional): unipolar or bipolar montage. Defaults to MontageType.UNIPOLAR.
            signalHeader (SignalHeader, optional): metadata of an EEG channel as defined by pyedflib.
                                                   Defaults to DEFAULT_SIGNAL_HEADER.
            fileHeader (FileHeader, optional): Metadata of an EEG file as defined by pyedflib. Defaults to DEFAULT_FILE_HEADER.
        """
        self.data = data
        self.channels = channels
        self.fs = fs
        self.montage = montage
        self._signalHeader = signalHeader
        self._fileHeader = fileHeader

    @classmethod
    def loadEdf(
        cls,
        edfFile: str,
        montage: Montage = Montage.UNIPOLAR,
        electrodes: list[str] = ELECTRODES_10_20,
    ):
        """Instantiate an Eeg object from an EDF file.

        Args:
            edfFile (str): path to EDF file.
            montage (Montage, optional): montage of the EEG recording. Defaults to Montage.UNIPOLAR.
            electrodes (list[str], optional): electrodes to load. If None all electrodes are loaded.
                                              For a bipolar montage, electrodes are expected in dash separated pairs
                                              (e.g. Fp1-F3). Defaults to the 19 electrodes of the 10-20 system.

        Returns:
            Eeg: returns an Eeg instance containing the data of the EDF file.
        """
        with pyedflib.EdfReader(edfFile) as edf:
            samplingFrequencies = edf.getSampleFrequencies()
            data = list()
            channels = list()
            # If electrodes are provided, load them
            if electrodes is not None:
                allChannels = edf.getSignalLabels()
                for electrode in electrodes:
                    try:
                        index = Eeg._findChannelIndex(allChannels, electrode, montage)
                        data.append(edf.readSignal(index))
                        channels.append(edf.getLabel(index))
                    except ValueError:
                        print(
                            f"Missing electrode {electrode} in file {edfFile} replaced by zeros."
                        )
                        data.append(np.zeros_like(edf.readSignal(0)))
                        channels.append(electrode)
            # Else read all channels with a fixed fs
            else:
                index = 0
                fixedFs = samplingFrequencies[index]
                for i, fs in enumerate(samplingFrequencies):
                    if fixedFs == fs:
                        data.append(edf.readSignal(i))
                        channels.append(edf.getLabel(i))
            signalHeader = edf.getSignalHeader(index)
            fileHeader = edf.getHeader()
            edf._close()

        return cls(
            np.array(data),
            channels,
            samplingFrequencies[index],
            montage,
            signalHeader,
            fileHeader,
        )

    @classmethod
    def loadEdfAutoDetectMontage(cls, edfFile: str):
        """Instantiate an Eeg object from an EDF file while auto-detecting electrodes and montage.

        Args:
            edfFile (str): path to EDF file.

        Returns:
            Eeg: returns an Eeg instance containing the data of the EDF file.
        """
        with pyedflib.EdfReader(edfFile) as edf:
            channel = edf.getLabel(0)
            edf._close()
        if channel == Eeg.ELECTRODES_10_20[0].upper():
            montage = Eeg.Montage.UNIPOLAR
            electrodes = Eeg.ELECTRODES_10_20
        elif channel == Eeg.BIPOLAR_DBANANA[0].upper():
            montage = Eeg.Montage.BIPOLAR
            electrodes = Eeg.BIPOLAR_DBANANA
        else:
            raise ValueError(
                f"Unrecognized electrode: {channel}. Expected {Eeg.ELECTRODES_10_20[0]} or {Eeg.BIPOLAR_DBANANA[0]}"
            )

        return cls.loadEdf(edfFile, montage, electrodes)

    def resample(self, newFs: int):
        """Resample data to a new sampling frequency.

        Args:
            newFs (int): new sampling frequency in Hz.
        """
        self.data = resampy.resample(self.data, self.fs, newFs)
        self.fs = newFs

    def reReferenceToBipolar(self):
        """Rereference unipolar data to a double-banana bipolar montage.

        Raises:
            TypeError: raised if Eeg data is not in a unipolar montage.
        """
        if self.montage is not Eeg.Montage.UNIPOLAR:
            raise TypeError("Data must be unipolar to re-reference.")
        # Create re-referencing matrix
        reRefMatrix = np.zeros((len(Eeg.BIPOLAR_DBANANA), self.data.shape[0]))
        for i, pair in enumerate(Eeg.BIPOLAR_DBANANA):
            elecs = pair.split("-")
            for elec, multiplier in zip(elecs, (1, -1)):
                index = Eeg._findChannelIndex(self.channels, elec, self.montage)
                reRefMatrix[i, index] = multiplier

        # Rereference data
        self.data = reRefMatrix @ self.data  # matrix multiplication
        self.channels = Eeg.BIPOLAR_DBANANA
        self.montage = Eeg.Montage.BIPOLAR

    def reReferenceToCommonAverage(self):
        """Rereference unipolar data to a common average montage.

        Raises:
            TypeError: raised if Eeg data is not in a unipolar montage.
        """
        if self.montage is not Eeg.Montage.UNIPOLAR:
            raise TypeError("Data must be unipolar to re-reference.")
        self.data -= np.mean(self.data, axis=0)
        self._constructUnipolarChannelNames("Avg")

    def reReferenceToReferential(self, electrode):
        """Rereference unipolar data to a given electrode.

        Raises:
            TypeError: raised if Eeg data is not in a unipolar montage.
        """
        if self.montage is not Eeg.Montage.UNIPOLAR:
            raise TypeError("Data must be unipolar to re-reference.")
        index = Eeg._findChannelIndex(self.channels, electrode, self.montage)
        self.data -= self.data[index]
        self._constructUnipolarChannelNames(electrode)

    def standardize(
        self,
        fs: int = 256,
        electrodes: list[str] = ELECTRODES_10_20,
        reference: str = "Avg",
    ):
        """Standardize data to a given sampling frequency, with a given set of electrodes and a given reference.

        Args:
            fs (int, optional): sampling frequency in Hz. Defaults to 256.
            electrodes (list[str], optional): set of electrodes to keep. If None all electrodes are loaded.
                                              For a bipolar montage, electrodes are expected in dash separated pairs
                                              (e.g. Fp1-F3). Defaults to the 19 electrodes of the 10-20 system.
            reference (str, optional): string representing referencing schemes. Supported values are:
                                       "Avg": common average unipolar montage.
                                       "bipolar:" double banana bipolar montage.
                                       electrode: name of the reference electrode for a unipolar referential montage.
                                       Defaults to "Avg".

        Raises:
            ValueError: raised if referencing scheme is unknown
        """
        # Select electrodes
        if electrodes is not None:
            reRefMatrix = np.zeros((len(electrodes), self.data.shape[0]))
            for i, electrode in enumerate(electrodes):
                index = Eeg._findChannelIndex(self.channels, electrode, self.montage)
                reRefMatrix[i, index] = 1
            # Select data
            self.data = reRefMatrix @ self.data
            # Select channels
            indices = np.where(reRefMatrix)[1]
            self.channels = [self.channels[i] for i in indices]

        # Resample
        self.resample(fs)

        # Re-Reference
        if reference == "Avg":
            self.reReferenceToCommonAverage()
        elif reference in electrodes:
            self.reReferenceToReferential(reference)
        elif reference == "bipolar":
            # Currently we trust bipolar montage without re-referencing
            # TODO attempt to re-reference bipolar montage if possible
            if self.montage is not Eeg.Montage.BIPOLAR:
                self.reReferenceToBipolar()
        else:
            raise ValueError("Unknown referencing scheme: {}".format(reference))

    def saveEdf(self, file: str):
        """Save Eeg object to an EDF file.

        Args:
            file (str): path of the file to save to. If directory does not exist it is created.
        """
        signalHeaders = list()
        for i, channel in enumerate(self.channels):
            signalHeaders.append(self._signalHeader.copy())
            signalHeaders[i]["label"] = channel
            signalHeaders[i]["sample_frequency"] = self.fs
            signalHeaders[i]["physical_min"] = int(np.floor(np.min(self.data[i])))
            signalHeaders[i]["physical_max"] = int(np.ceil(np.max(self.data[i])))

        # Create directory for file
        if os.path.dirname(file):
            os.makedirs(os.path.dirname(file), exist_ok=True)
        # Write new EDF file
        pyedflib.highlevel.write_edf(file, self.data, signalHeaders, self._fileHeader)

    def saveDataFrame(self, file: str, format: FileFormat = FileFormat.PARQUET_GZIP):
        """Save Eeg object to a dataframe compatible file.

        Args:
            file (str):  path of the file to save to. If directory does not exist it is created.
            format (FileFormat, optional): File format to save to. Defaults to FileFormat.PARQUET_GZIP.

        Raises:
            ValueError: raised if fileFormat is not supported.
        """
        tmpData = self.data.copy()
        dataDF = pd.DataFrame(data=tmpData.transpose(), columns=self.channels)
        # TODO save metadata -- pyarrow might be a good candidate
        dataDF.attrs["fileHeader"] = self._fileHeader
        dataDF.attrs["fileHeader"]["startdate"] = str(
            dataDF.attrs["fileHeader"]["startdate"]
        )
        dataDF.attrs["signalHeader"] = self._signalHeader

        # Create directory for file
        if os.path.dirname(file):
            os.makedirs(os.path.dirname(file), exist_ok=True)
        # Write new file
        match format:
            case FileFormat.PARQUET_GZIP:
                dataDF.to_parquet(file, index=False, compression="gzip")
            case FileFormat.CSV_GZIP:
                dataDF.to_csv(file, index=False, compression="gzip")
            case FileFormat.CSV:
                dataDF.to_csv(file, index=False)
            case _:
                raise ValueError("Unknown output format {}".format(format))

    def _electrodeSynonymRegex(electrode: str) -> str:
        """Build a regex that matches the different synonyms of an electrode name.

        Args:
            electrode (str): electrode name

        Returns:
            str: regex that matches the different synonyms of an electrode name
        """
        if electrode == "T3" or electrode == "T7":
            electrode = "(T3|T7)"
        elif electrode == "T4" or electrode == "T8":
            electrode = "(T4|T8)"
        elif electrode == "T5" or electrode == "P7":
            electrode = "(T5|P7)"
        elif electrode == "T6" or electrode == "P8":
            electrode = "(T6|P8)"
        elif electrode == "O1":
            electrode = "(O1|01)"

        return electrode

    def _findChannelIndex(
        channels: tuple[str], electrode: str, montage: Montage
    ) -> int:
        """Finds the index of a channel given an electrode name.

        Args:
            channels (tuple[str]): list of channel names
            electrode (str): electrode to search for. For a bipolar montage, electrodes are expected in dash separated pairs.
            montage (Montage): montage of the data (unipolar or bipolar)

        Raises:
            ValueError: raised if electrode is not found in the list of channels

        Returns:
            int: index of the electrode in the channels array
        """
        # Some channel have different names in different datasets.
        if montage == Eeg.Montage.UNIPOLAR:
            electrode = Eeg._electrodeSynonymRegex(electrode)
        elif montage == Eeg.Montage.BIPOLAR:
            electrodes = electrode.split("-")
            for i, elec in enumerate(electrodes):
                electrodes[i] = Eeg._electrodeSynonymRegex(elec)

        # Regex on channel name
        if montage == Eeg.Montage.UNIPOLAR:
            regExToFind = r"^(EEG )?{}(-[a-z]?[1-9]*)?".format(electrode)
        elif montage == Eeg.Montage.BIPOLAR:
            regExToFind = electrodes[0] + r".*(-)?.*" + electrodes[1]

        # Find corresponding channel
        index = None
        for i, channel in enumerate(channels):
            if re.search(regExToFind, channel, flags=re.IGNORECASE):
                index = i
                break
        if index is None:
            raise ValueError("Electrode {} was not found in EDF file".format(electrode))

        return index

    def _constructUnipolarChannelNames(self, reference: str = "REF"):
        """Rename channels to a standardized name of the format ELEC-REF.

        Args:
            reference (str, optional): reference. Defaults to "REF".
        """
        regExToFind = r"^(EEG )?([A-Z]{1,2}[1-9]*)(-[a-z]?[1-9]*)?"
        for i, channel in enumerate(self.channels):
            result = re.search(regExToFind, channel, flags=re.IGNORECASE)
            if result.group(2) is not None:
                electrode = result.group(2)
            else:
                electrode = channel
            self.channels[i] = "{}-{}".format(electrode, reference)



import os
from importlib import resources as impresources
from pathlib import Path
from string import Template

import pandas as pd

from ... import bids
from ...bids.convert2bids import BidsConverter
from ...eeg import Eeg
from ...load_annotations.chbmit import loadAnnotationsFromEdf

BIDS_DIR = impresources.files(bids)
DATASET = BIDS_DIR / "chbmit"


def convert(root: Path, outDir: Path):
    root = Path(root)
    outDir = Path(outDir)
    bidsConverter = BidsConverter(BIDS_DIR, DATASET, root, outDir, loadAnnotationsFromEdf)
    subjects = []
    for _, directory, _ in os.walk(root):
        for subject in directory:
            subjects.append(subject)
    for folder in subjects:
        print(folder)
        # Extract subject & session ID
        subject = os.path.split(folder)[-1][3:5]
        session = "01"
        task = "szMonitoring"
        if subject == "21":
            subject = "01"
            session = "02"

        # Create BIDS hierarchy
        outPath = outDir / f"sub-{subject}" / f"ses-{session}" / "eeg"
        os.makedirs(outPath, exist_ok=True)

        edfFiles = sorted((root / folder).glob("*.edf"))
        for fileIndex, edfFile in enumerate(edfFiles):
            edfBaseName = (
                outPath
                / f"sub-{subject}_ses-{session}_task-{task}_run-{fileIndex:02}_eeg"
            )
            edfFileName = edfBaseName.with_suffix(".edf")
            # Load EEG and standardize it
            if os.path.basename(edfFile) not in (
                "chb12_27.edf",
                "chb12_28.edf",
                "chb12_29.edf",
            ):
                eeg = Eeg.loadEdf(
                    edfFile.as_posix(), Eeg.Montage.BIPOLAR, Eeg.BIPOLAR_DBANANA
                )
                eeg.standardize(256, Eeg.BIPOLAR_DBANANA, "bipolar")
            else:
                eeg = Eeg.loadEdf(
                    edfFile.as_posix(), Eeg.Montage.UNIPOLAR, Eeg.ELECTRODES_10_20
                )
                eeg.standardize(256, Eeg.ELECTRODES_10_20, "bipolar")

            # Save EEG
            eeg.saveEdf(edfFileName.as_posix())

            # Save JSON sidecar
            eegJsonDict = {
                "fs": f"{eeg.fs:d}",
                "channels": f"{eeg.data.shape[0]}",
                "duration": f"{(eeg.data.shape[1] / eeg.fs):.2f}",
                "task": task,
            }

            with open(DATASET / "eeg.json", "r") as f:
                src = Template(f.read())
                eegJsonSidecar = src.substitute(eegJsonDict)
            with open(edfBaseName.with_suffix(".json"), "w") as f:
                f.write(eegJsonSidecar)

            # Load annotation
            annotations = loadAnnotationsFromEdf(edfFile.as_posix())
            annotations.saveTsv(edfBaseName.as_posix()[:-4] + "_events.tsv")

    # Build participant metadata
    subjectInfo = pd.read_csv(
        root / "SUBJECT-INFO", delimiter="\t", skip_blank_lines=True
    )
    participants = {"participant_id": [], "age": [], "sex": [], "comment": []}
    for folder in outDir.glob("sub-*"):
        subject = os.path.split(folder)[-1]
        originalSubjectName = f"chb{subject[4:6]}"
        if originalSubjectName in subjectInfo.Case.values:
            participants["participant_id"].append(subject)
            participants["age"].append(
                subjectInfo[subjectInfo.Case == originalSubjectName][
                    "Age (years)"
                ].values[0]
            )
            participants["sex"].append(
                subjectInfo[subjectInfo.Case == originalSubjectName]
                .Gender.values[0]
                .lower()
            )
            if subject == "sub-01":
                participants["comment"].append("ses-02 recorded at 13 years old")
            else:
                participants["comment"].append("n/a")

        else:
            participants["participant_id"].append(subject)
            participants["age"].append("n/a")
            participants["sex"].append("n/a")
            participants["comment"].append("n/a")

    bidsConverter.saveMetadata(participants)



import os
import shutil
from string import Template

import pandas as pd

from ..eeg import Eeg

class BidsConverter:
    def __init__(self, BIDS_DIR, DATASET, root, outDir, loadAnnotationsFromEdf, montage = Eeg.Montage.UNIPOLAR, electrodes = Eeg.ELECTRODES_10_20):
        self.BIDS_DIR = BIDS_DIR
        self.DATASET = DATASET
        self.root = root
        self.outDir = outDir
        self.loadAnnotationsFromEdf = loadAnnotationsFromEdf
        self.montage = montage
        self.electrodes = electrodes


    def buildBIDSHierarchy(self, edfFiles, subject, session = "01", task = "szMonitoring", addEegJsonDict = None):        
        # Create BIDS hierarchy
        outPath = self.outDir / f"sub-{subject}" / f"ses-{session}" / "eeg"
        os.makedirs(outPath, exist_ok=True)
        for fileIndex, edfFile in enumerate(edfFiles):
            edfBaseName = (
                outPath / f"sub-{subject}_ses-{session}_task-{task}_run-{(fileIndex + 1):02}_eeg"
            )
            edfFileName = edfBaseName.with_suffix(".edf")
            # Load EEG and standardize it
            eeg = Eeg.loadEdf(edfFile.as_posix(), self.montage, self.electrodes)
            eeg.standardize(256, self.electrodes, "Avg")

            # Save EEG
            eeg.saveEdf(edfFileName.as_posix())

            # Save JSON sidecar
            eegJsonDict = {
                "fs": f"{eeg.fs:d}",
                "channels": f"{eeg.data.shape[0]}",
                "duration": f"{(eeg.data.shape[1] / eeg.fs):.2f}",
                "task": task,
            }
            if addEegJsonDict is not None:
                eegJsonDict = eegJsonDict | addEegJsonDict

            with open(self.DATASET / "eeg.json", "r") as f:
                src = Template(f.read())
                eegJsonSidecar = src.substitute(eegJsonDict)
            with open(edfBaseName.with_suffix(".json"), "w") as f:
                f.write(eegJsonSidecar)

            # Load annotation
            annotations = self.loadAnnotationsFromEdf(edfFile.as_posix())
            annotations.saveTsv(edfBaseName.as_posix()[:-4] + "_events.tsv")


    def saveMetadata(self, participants):
        participantsDf = pd.DataFrame(participants)
        participantsDf.sort_values(by=["participant_id"], inplace=True)
        participantsDf.to_csv(self.outDir / "participants.tsv", sep="\t", index=False)
        participantsJsonFileName = self.DATASET / "participants.json"
        shutil.copy(participantsJsonFileName, self.outDir)

        # Copy Readme file
        readmeFileName = self.DATASET / "README.md"
        shutil.copyfile(readmeFileName, self.outDir / "README")

        # Copy dataset description
        descriptionFileName = self.DATASET / "dataset_description.json"
        shutil.copy(descriptionFileName, self.outDir)

        # Copy Events JSON Sidecar
        eventsFileName = self.BIDS_DIR / "events.json"
        shutil.copy(eventsFileName, self.outDir)



import enum
import json
from datetime import datetime
from importlib import resources as impresources
from typing import List, Tuple, TypedDict

import numpy as np
import pandas as pd
from timescoring.annotations import Annotation as MaskEvents

from . import bids

# Load Seizure types defined in the HED-SCORE JSON event file
BIDS_LOC = impresources.files(bids)
szTypes = dict()
with open(BIDS_LOC / "events.json", "r") as f:
    eventsJSON = json.load(f)
    szTypes = eventsJSON["Levels"]

for key, _ in szTypes.items():
    if key != "bckg":
        szTypes[key] = key

SeizureType = enum.Enum("SeizureType", szTypes)

EVENT_TYPES = szTypes.copy()
EVENT_TYPES["bckg"] = "bckg"
EventType = enum.Enum("EventType", EVENT_TYPES)

# TODO Subclass dataFrame
# Q? Expose dataFrame or getter class to get masks and annotations


class Annotation(TypedDict):
    onset: (
        float  # start time of the event from the beginning of the recording, in seconds
    )
    duration: float  # duration of the event, in seconds
    eventType: EventType  # type of the event
    confidence: float  # confidence in the event label. Values are in the range [0–1]
    channels: list[str]  # channels on which the event appears
    dateTime: datetime  # start date time of the recording file
    recordingDuration: float  # duration of the recording in seconds


class Annotations:
    def __init__(self):
        self.events: list[Annotation] = list()

    @classmethod
    def loadTsv(cls, filename: str):
        df = pd.read_csv(filename, delimiter="\t")
        annotations = cls()
        for _, row in df.iterrows():
            annotation = Annotation()
            annotation["onset"] = float(row["onset"])
            annotation["duration"] = float(row["duration"])
            annotation["eventType"] = EventType[row["eventType"]]
            if row["confidence"]:
                annotation["confidence"] = "n/a"
            else:
                annotation["confidence"] = float(row["confidence"])
            if row["channels"]:
                annotation["channels"] = "n/a"
            elif "," in row["channels"]:
                annotation["channels"] = row["channels"].split(",")
            else:
                annotation["channels"] = [row["channels"]]
            annotation["dateTime"] = datetime.strptime(
                row["dateTime"], "%Y-%m-%d %H:%M:%S"
            )
            annotation["recordingDuration"] = float(row["recordingDuration"])
            annotations.events.append(annotation)

        return annotations

    @classmethod
    def loadMask(cls, mask, fs):
        maskEvent = MaskEvents(mask, fs)
        return cls.loadEvents(maskEvent.events, len(mask) / fs)

    @classmethod
    def loadEvents(cls, events: List[Tuple[float, float]], duration: float):
        annotations = cls()
        for event in events:
            annotation = Annotation()
            annotation["onset"] = event[0]
            annotation["duration"] = event[1] - event[0]
            annotation["eventType"] = SeizureType.sz
            annotation["confidence"] = "n/a"
            annotation["channels"] = "n/a"
            annotation["dateTime"] = "n/a"
            annotation["recordingDuration"] = duration
            annotations.events.append(annotation)
        if len(events) == 0:
            annotation = Annotation()
            annotation["onset"] = 0
            annotation["duration"] = duration
            annotation["eventType"] = EventType.bckg
            annotation["confidence"] = "n/a"
            annotation["channels"] = "n/a"
            annotation["dateTime"] = "n/a"
            annotation["recordingDuration"] = duration
            annotations.events.append(annotation)
        return annotations

    def getEvents(self) -> list[(float, float)]:
        events = list()
        for event in self.events:
            if event["eventType"].value in SeizureType._member_names_:
                events.append((event["onset"], event["onset"] + event["duration"]))
        return events

    def getMask(self, fs: int) -> np.ndarray:
        mask = np.zeros(int(self.events[0]["recordingDuration"] * fs))
        for event in self.events:
            if event["eventType"].value in SeizureType._member_names_:
                mask[
                    int(event["onset"] * fs) : int(
                        (event["onset"] + event["duration"]) * fs
                    )
                ] = 1
        return mask

    def saveTsv(self, filename: str):
        with open(filename, "w") as f:
            line = "\t".join(list(Annotation.__annotations__.keys()))
            line += "\n"
            f.write(line)
            for event in self.events:
                line = ""
                line += "{:.2f}\t".format(event["onset"])
                line += "{:.2f}\t".format(event["duration"])
                line += "{}\t".format(event["eventType"].value)
                if isinstance(event["confidence"], (int, float)):
                    line += "{:.2f}\t".format(event["confidence"])
                else:
                    line += "{}\t".format(event["confidence"])
                if isinstance(event["channels"], (list, tuple)):
                    line += ",".join(event["channels"])
                    line += "\t"
                else:
                    line += "{}\t".format(event["channels"])
                if isinstance(event["dateTime"], datetime):
                    line += "{}\t".format(event["dateTime"].strftime("%Y-%m-%d %H:%M:%S"))
                else:
                    line += "{}\t".format(event["dateTime"])
                line += "{:.2f}".format(event["recordingDuration"])
                line += "\n"
                f.write(line)



import os
from importlib import resources as impresources
from pathlib import Path

import pandas as pd

from ... import bids
from ...bids.convert2bids import BidsConverter
from ...load_annotations.siena import loadAnnotationsFromEdf

BIDS_DIR = impresources.files(bids)
DATASET = BIDS_DIR / "siena"


def convert(root: Path, outDir: Path):
    root = Path(root)
    outDir = Path(outDir)
    bidsConverter = BidsConverter(BIDS_DIR, DATASET, root, outDir, loadAnnotationsFromEdf)

    for folder in root.glob("PN*"):
        print(folder)
        # Extract subject & session ID
        subject = folder.name[-2:]
        edfFiles = sorted((root / folder).glob("*.edf"))
        bidsConverter.buildBIDSHierarchy(edfFiles, subject)

    # Build participant metadata
    subjectInfo = pd.read_csv(root / "subject_info.csv")
    participants = {"participant_id": [], "age": [], "sex": []}
    for folder in outDir.glob("sub-*"):
        subject = os.path.split(folder)[-1]
        originalSubjectName = f"PN{subject[-2:]}"
        if originalSubjectName in subjectInfo.patient_id.values:
            participants["participant_id"].append(subject)
            participants["age"].append(
                subjectInfo[subjectInfo.patient_id == originalSubjectName][
                    " age_years"
                ].values[0]
            )
            participants["sex"].append(
                subjectInfo[subjectInfo.patient_id == originalSubjectName][" gender"]
                .values[0]
                .lower()[0]
            )

        else:
            participants["participant_id"].append(subject)
            participants["age"].append("n/a")
            participants["sex"].append("n/a")

    bidsConverter.saveMetadata(participants)



import os
from importlib import resources as impresources
from pathlib import Path

from ... import bids
from ...bids.convert2bids import BidsConverter
from ...load_annotations.seizeit import loadAnnotationsFromEdf

BIDS_DIR = impresources.files(bids)
DATASET = BIDS_DIR / "seizeit"


def convert(root: Path, outDir: Path):
    root = Path(root)
    outDir = Path(outDir)
    bidsConverter = BidsConverter(BIDS_DIR, DATASET, root, outDir, loadAnnotationsFromEdf)
    for folder in root.glob("P_ID*"):
        print(folder)
        # Extract subject & session ID
        subject = folder.name[-2:]
        edfFiles = sorted((root / folder).glob("*.edf"))
        bidsConverter.buildBIDSHierarchy(edfFiles, subject)

    # Build participant metadata
    participants = {"participant_id": [], "age": [], "sex": []}
    for folder in outDir.glob("P_ID*"):
        print(folder)
        subject = os.path.split(folder)[-1]
        originalSubjectName = f"P_ID{subject[-2:]}"
        # Get information from header in one of the metadata files
        annotationFile = next((root / originalSubjectName).glob("*_a1.tsv"), None)
        with open(annotationFile, "r") as f:
            lines = f.readlines()
            age = int(lines[4].split(": ")[-1])
            sex = lines[3].split(": ")[-1][0].lower()
        participants["participant_id"].append(subject)
        participants["age"].append(age)
        participants["sex"].append(sex)

    bidsConverter.saveMetadata(participants)



"""load annotations from the SeizeIT dataset https://doi.org/10.48804/P5Q0OJ to a Annotations object."""

import math
import os
import re
from pathlib import Path

import pandas as pd
import pyedflib

from ..annotations import Annotation, Annotations, EventType, SeizureType
from ..eeg import Eeg


def _loadSeizures(
    edfFile: str,
) -> tuple[list[tuple], list[SeizureType], list[list[str]], list]:
    """Load seizures from a seizeIT .tsv file

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Raises:
        ValueError: raises a ValueError if txt file format is unknown.

    Returns:
        seizures: A list of (start, end) tuples for each seizure in seconds from the beginning of the file.
        types: A list off seizureType
        confidence: A list of confidence for each seizure. The value is 0.5 for the 12 seizures that do not have an endt-time.
        channels: A list of channels the seizure is visible on for each seizure.
    """
    seizures = []
    types = []
    confidence = []
    channels = []
    tsvFile = Path(os.path.dirname(edfFile)) / (Path(edfFile).stem + "_a1.tsv")
    annotations = pd.read_csv(
        tsvFile,
        comment="#",
        delimiter="\t",
        names=["start", "stop", "type", "comments"],
    )
    for _, row in annotations.iterrows():
        # Seizure Timing
        if not math.isnan(row["stop"]):
            seizures.append((row["start"], row["stop"]))
            confidence.append("n/a")
        # 12 weizures do not mark the end time, consider a default seizure time of 30 seconds
        else:
            seizures.append((row["start"], row["start"] + 30))
            confidence.append(0.5)

        # Seizure Type
        seizureType = row["type"]
        if seizureType == "FIA":
            seizureType = SeizureType.sz_foc_ia
        elif seizureType == "FA":
            seizureType = SeizureType.sz_foc_a
        elif seizureType == "F-BTC":
            seizureType = SeizureType.sz_foc_f2b
        else:
            raise ValueError(f"Unknown seizure type ({seizureType}) for Siena dataset.")
        types.append(seizureType)

        # Seizure localization
        channels.append(_getChannels(row["comments"]))

    return seizures, types, confidence, channels


def _getChannels(descriptor: str) -> list[str]:
    """Get channels associated with a seizure from the channel description.

    Args:
        descriptor (str): SeizeIT channel description

    Raises:
        ValueError: raises a ValueError if the lateralization or localization is unknown.

    Returns:
        channels: List of electrodes names from the 10-20 system associated associated with a seizure.
    """
    localization = descriptor.split(", ")[1].split(":")[1].lower()
    lateralization = descriptor.split(", ")[0].split(":")[1]
    channels = Eeg.ELECTRODES_10_20
    r = list()
    if "temp" in localization:  # Temporal
        r.append(re.compile("^T.*"))
    if "fronto" in localization:  # Frontal
        r.append(re.compile("^F.*"))
    if "occipito" in localization:  # Occipital
        r.append(re.compile("^O.*"))
    if "par" in localization:  # Occipital
        r.append(re.compile("^P.*"))
    if localization == "NC":
        r = None
    if r is not None:
        channels = []
        for regex in r:
            channels += list(filter(regex.match, Eeg.ELECTRODES_10_20))
    else:
        channels = "n/a"

    match lateralization:
        case "R":
            r = re.compile(".*[02468]$")
        case "L":
            r = re.compile(".*[13579]$")
        case "bi":
            r = re.compile(".*")
        case "NC":
            r = None
        case _:
            raise ValueError(
                f"Unknown lateralization for SeizeIT dataset: {lateralization}."
            )
    if r is not None:
        channels = list(filter(r.match, channels))
    else:
        channels = "n/a"

    return channels


def loadAnnotationsFromEdf(edfFile: str) -> Annotations:
    """Loads annotations related to an EDF recording in the Siena dataset.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Returns:
        Annotations: an Annotations object
    """
    # dateTime and duration
    with pyedflib.EdfReader(edfFile) as edf:
        dateTime = edf.getStartdatetime()
        duration = edf.getFileDuration()
        edf._close()

    # Load event file
    seizures, types, confidence, channels = _loadSeizures(edfFile)

    # Populate dictionary
    if len(seizures) == 0:
        types = [EventType.bckg]
        confidence = ["n/a"]
        channels = ["n/a"]
        seizures.append((0, duration))

    annotations = Annotations()
    for i, seizure in enumerate(seizures):
        annotation = Annotation()
        annotation["onset"] = seizure[0]
        annotation["duration"] = seizure[1] - seizure[0]
        annotation["eventType"] = types[i]
        annotation["confidence"] = confidence[i]
        annotation["channels"] = channels[i]
        annotation["dateTime"] = dateTime
        annotation["recordingDuration"] = duration
        annotations.events.append(annotation)

    return annotations



"""load annotations from the TUH Sz Corpus dataset https://isip.piconepress.com/projects/tuh_eeg/downloads/tuh_eeg_seizure/ to a Annotations object."""

import os
from pathlib import Path

import pandas as pd
import pyedflib

from ..annotations import Annotation, Annotations, EventType, SeizureType


def _loadSeizures(
    edfFile: str,
) -> tuple[list[tuple], list[SeizureType], list[list[str]], list]:
    """Load seizures from a TUH .csv_bi file

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Raises:
        ValueError: raises a ValueError if txt file format is unknown.

    Returns:
        seizures: A list of (start, end) tuples for each seizure in seconds from the beginning of the file.
        types: A list off seizureType
        confidence: A list of confidence for each seizure. The value is 0.5 for the 12 seizures that do not have an endt-time.
        channels: A list of channels the seizure is visible on for each seizure.
    """

    MAPPING = {
        "BCKG": EventType.bckg,
        "SEIZ": SeizureType.sz,
        "FNSZ": SeizureType.sz_foc,
        "GNSZ": SeizureType.sz_gen,
        "SPSZ": SeizureType.sz_foc_a,
        "CPSZ": SeizureType.sz_foc_ia,
        "ABSZ": SeizureType.sz_gen_nm,
        "TNSZ": SeizureType.sz_gen_m_tonic,
        "CNSZ": SeizureType.sz_gen_m_clonic,
        "TCSZ": SeizureType.sz_gen_m_tonicClonic,
        "ATSZ": SeizureType.sz_gen_m_atonic,
        "MYSZ": SeizureType.sz_gen_nm_myoclonic,
    }
    seizures = []
    types = []
    confidence = []
    channels = []
    csvFile = Path(os.path.dirname(edfFile)) / (Path(edfFile).stem + ".csv_bi")
    annotations = pd.read_csv(csvFile, comment="#", delimiter=",")
    for _, row in annotations.iterrows():
        # Seizure Timing
        seizures.append((row["start_time"], row["stop_time"]))
        types.append(MAPPING[row["label"].upper()])
        confidence.append(float(row["confidence"]))  # TODO

        # Seizure localization
        if row["channel"] == "TERM":
            channels.append("n/a")
        else:
            raise ValueError(f"Unknown channel: {row['channel']}")

    return seizures, types, confidence, channels


def loadAnnotationsFromEdf(edfFile: str) -> Annotations:
    """Loads annotations related to an EDF recording in the Siena dataset.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Returns:
        Annotations: an Annotations object
    """
    # dateTime and duration
    with pyedflib.EdfReader(edfFile) as edf:
        dateTime = edf.getStartdatetime()
        duration = edf.getFileDuration()
        edf._close()

    # Load event file
    seizures, types, confidence, channels = _loadSeizures(edfFile)

    # Populate dictionary
    if len(seizures) == 0:
        types = [EventType.bckg]
        confidence = ["n/a"]
        channels = ["n/a"]
        seizures.append((0, duration))

    annotations = Annotations()
    for i, seizure in enumerate(seizures):
        annotation = Annotation()
        annotation["onset"] = seizure[0]
        annotation["duration"] = seizure[1] - seizure[0]
        annotation["eventType"] = types[i]
        annotation["confidence"] = confidence[i]
        annotation["channels"] = channels[i]
        annotation["dateTime"] = dateTime
        annotation["recordingDuration"] = duration
        annotations.events.append(annotation)

    return annotations



"""Script to build a small sample EDF from an EDF recording. The recording header and list of signals are kept the same.
The content of data is white noise. The output file is named input_sample.edf
"""

import argparse
from itertools import islice
import os
from pathlib import Path
import shutil

from makeSampleEdf import makeSample

DURATION = 2
AMPLITUDE = 100


def makeSampleChb(input: Path, output: Path, numSubjects: int = 2) -> None:
    output.mkdir(parents=True, exist_ok=True)

    # Copy subject infor
    shutil.copy(input / "SUBJECT-INFO", output)

    folders = input.glob("*/")
    for folder in islice(folders, numSubjects - 1):
        subject = folder.name

        # Copy sz summary
        (output / subject).mkdir(parents=True, exist_ok=True)
        shutil.copy(folder / f"{subject}-summary.txt", output / subject)

        # Copy truncated EDF recordings
        for edfFile in folder.glob("*.edf"):
            makeSample(edfFile, output / subject / edfFile.name)


def makeSampleSiena(input: Path, output: Path, numSubjects: int = 2) -> None:
    output.mkdir(parents=True, exist_ok=True)

    # Copy subject info
    shutil.copy(input / "subject_info.csv", output)

    folders = input.glob("PN*/")
    for folder in islice(folders, numSubjects - 1):
        subject = folder.name

        # Copy sz summary
        (output / subject).mkdir(parents=True, exist_ok=True)
        shutil.copy(folder / f"Seizures-list-{subject}.txt", output / subject)

        # Copy truncated EDF recordings
        for edfFile in folder.glob("*.edf"):
            makeSample(edfFile, output / subject / edfFile.name)


def makeSampleSeizeit(input: Path, output: Path, numSubjects: int = 2) -> None:
    output.mkdir(parents=True, exist_ok=True)

    folders = input.glob("*/")
    for folder in islice(folders, numSubjects - 1):
        subject = folder.name
        (output / subject).mkdir(parents=True, exist_ok=True)
        # Copy truncated EDF recordings and annotations
        for edfFile in folder.glob("*.edf"):
            makeSample(edfFile, output / subject / edfFile.name)
            shutil.copy(
                edfFile.with_stem(f"{edfFile.stem}_a1").with_suffix(".tsv"),
                output / subject,
            )


def makeSampleTuh(input: Path, output: Path, numSubjects: int = 2) -> None:
    for subset in ["train", "dev", "eval"]:
        subRoot = input / subset
        folders = subRoot.glob("*/")
        for folder in islice(folders, numSubjects - 1):                    
            for edfFile in folder.glob("**/*.edf"):
                (output / os.path.join(*edfFile.parts[-5:-1])).mkdir(parents=True, exist_ok=True)
                print(edfFile)
                makeSample(
                    edfFile, output / os.path.join(*edfFile.parts[-5:])
                )
                shutil.copy(
                    edfFile.with_suffix(".csv_bi"),
                    output / os.path.join(*edfFile.parts[-5:-1]),
                )

        subject = folder.name
        (output / subject).mkdir(parents=True, exist_ok=True)
        # Copy truncated EDF recordings and annotations
        for edfFile in folder.glob("*.edf"):
            makeSample(edfFile, output / subject / edfFile.name)
            shutil.copy(
                edfFile.with_stem(f"{edfFile.stem}_a1").with_suffix(".tsv"),
                output / subject,
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Extract an anonymized sample from a dataset",
        description="Shrinks an EDF to a sample. The recording header and list of signals are kept the same. "
        "The content of data is white noise. The output file is named input_sample.edf",
    )
    parser.add_argument("input", help="input folder.")
    parser.add_argument("output", help="output folder.")
    parser.add_argument("dataset", help="chbmit, siena, seizeit, tuh")

    args = parser.parse_args()
    match args.dataset:
        case "chbmit":
            makeSampleChb(Path(args.input), Path(args.output), 3)
        case "siena":
            makeSampleSiena(Path(args.input), Path(args.output), 3)
        case "seizeit":
            makeSampleSeizeit(Path(args.input), Path(args.output), 3)
        case "tuh":
            makeSampleTuh(Path(args.input), Path(args.output), 3)
        case _:
            raise ValueError(f"Unknown dataset: {args.dataset}.")



"""Eeg class unit testing"""

import copy
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from src.epilepsy2bids.eeg import Eeg, FileFormat


class TestDataLoading(unittest.TestCase):
    def test_loadEdf(self):
        fileConfigurations = [
            {  # CHB-MIT
                "fileName": "tests/chb01_01_sample.edf",
                "montage": Eeg.Montage.BIPOLAR,
                "electrodes": Eeg.BIPOLAR_DBANANA,
            },
            {  # TUH
                "fileName": "tests/aaaaaaac_s001_t000_sample.edf",
                "montage": Eeg.Montage.UNIPOLAR,
                "electrodes": Eeg.ELECTRODES_10_20,
            },
            {  # Siena
                "fileName": "tests/PN00-5_sample.edf",
                "montage": Eeg.Montage.UNIPOLAR,
                "electrodes": Eeg.ELECTRODES_10_20,
            },
            {  # SeizeIT
                "fileName": "tests/P_ID10_r5_sample.edf",
                "montage": Eeg.Montage.UNIPOLAR,
                "electrodes": Eeg.ELECTRODES_10_20,
            },
        ]

        for fileConfig in fileConfigurations:
            eeg = Eeg.loadEdf(
                fileConfig["fileName"], fileConfig["montage"], fileConfig["electrodes"]
            )
            self.assertEqual(eeg.data.shape[0], len(fileConfig["electrodes"]))
            self.assertEqual(len(eeg.channels), len(fileConfig["electrodes"]))
            self.assertEqual(eeg.montage, fileConfig["montage"])

    def test_resampling(self):
        fileConfig = {  # Siena
            "fileName": "tests/PN00-5_sample.edf",
            "montage": Eeg.Montage.UNIPOLAR,
            "electrodes": Eeg.ELECTRODES_10_20,
        }
        eeg = Eeg.loadEdf(
            fileConfig["fileName"], fileConfig["montage"], fileConfig["electrodes"]
        )
        fileDuration = eeg.data.shape[1] / eeg.fs
        newFs = 256
        eeg.resample(newFs)
        newFileDuration = eeg.data.shape[1] / newFs

        self.assertEqual(fileDuration, newFileDuration)
        self.assertEqual(eeg.fs, newFs)

    def test_reReference(self):
        fileConfig = {  # Siena
            "fileName": "tests/PN00-5_sample.edf",
            "montage": Eeg.Montage.UNIPOLAR,
            "electrodes": Eeg.ELECTRODES_10_20,
        }
        eeg = Eeg.loadEdf(
            fileConfig["fileName"], fileConfig["montage"], fileConfig["electrodes"]
        )
        # Common average
        eegAvg = copy.deepcopy(eeg)
        eegAvg.reReferenceToCommonAverage()
        # average of common average ref should be zero
        np.testing.assert_allclose(
            np.mean(eegAvg.data, axis=0),
            np.zeros((eeg.data.shape[1],)),
            rtol=1e-07,
            atol=1e-14,
        )
        # channel[0] should be data[0] - common average
        np.testing.assert_array_equal(
            eegAvg.data[0], eeg.data[0] - np.mean(eeg.data, axis=0)
        )
        # Cz Reference
        eegCz = copy.deepcopy(eeg)
        eegCz.reReferenceToReferential("Cz")
        # cz should be zero
        CzIndex = Eeg._findChannelIndex(eegCz.channels, "Cz", eegCz.montage)
        np.testing.assert_allclose(
            eegCz.data[CzIndex],
            np.zeros((eeg.data.shape[1],)),
            rtol=1e-07,
            atol=1e-14,
        )
        # Fz should be Fz - Cz
        FzIndex = Eeg._findChannelIndex(eegCz.channels, "Fz", eegCz.montage)
        np.testing.assert_array_equal(
            eegCz.data[FzIndex], eeg.data[FzIndex] - eeg.data[CzIndex]
        )
        # Check bipolar channel
        eegBp = copy.deepcopy(eeg)
        eegBp.reReferenceToBipolar()
        i0 = Eeg._findChannelIndex(
            eeg.channels, Eeg.BIPOLAR_DBANANA[0].split("-")[0], eeg.montage
        )
        i1 = Eeg._findChannelIndex(
            eeg.channels, Eeg.BIPOLAR_DBANANA[0].split("-")[1], eeg.montage
        )
        np.testing.assert_array_equal(eegBp.data[0], eeg.data[i0] - eeg.data[i1])

    def test_saveEdf(self):
        fileConfig = {  # Siena
            "fileName": "tests/PN00-5_sample.edf",
            "montage": Eeg.Montage.UNIPOLAR,
            "electrodes": Eeg.ELECTRODES_10_20,
        }
        eeg = Eeg.loadEdf(
            fileConfig["fileName"], fileConfig["montage"], fileConfig["electrodes"]
        )
        eeg.standardize()
        eeg.saveEdf("test.edf")

        standardEeg = Eeg.loadEdf(
            "test.edf", fileConfig["montage"], fileConfig["electrodes"]
        )
        self.assertEqual(standardEeg.fs, 256)
        self.assertListEqual(standardEeg.channels, eeg.channels)
        np.testing.assert_allclose(
            standardEeg.data, eeg.data, rtol=1e-7, atol=1e-2
        )  # TODO absolute error is high might need to be checked
        Path("test.edf").unlink()

    def test_savecsv(self):
        fileConfig = {  # Siena
            "fileName": "tests/PN00-5_sample.edf",
            "montage": Eeg.Montage.UNIPOLAR,
            "electrodes": Eeg.ELECTRODES_10_20,
        }
        eeg = Eeg.loadEdf(
            fileConfig["fileName"], fileConfig["montage"], fileConfig["electrodes"]
        )
        eeg.standardize()
        
        for ext in [FileFormat.CSV, FileFormat.CSV_GZIP, FileFormat.PARQUET_GZIP]:
            eeg.saveDataFrame(f"test.{ext}", ext)
            match ext:
                case FileFormat.CSV:
                    tblData = pd.read_csv(f"test.{ext}")
                case FileFormat.CSV_GZIP:
                    tblData = pd.read_csv(f"test.{ext}", compression="gzip")
                case FileFormat.PARQUET_GZIP:
                    tblData = pd.read_parquet(f"test.{ext}")
            self.assertListEqual(list(tblData), eeg.channels)
            np.testing.assert_allclose(eeg.data, tblData.to_numpy().transpose())
            Path(f"test.{ext}").unlink()


if __name__ == "__main__":
    unittest.main()



"""Annotation class unit testing"""

import unittest
from pathlib import Path

import numpy as np

from epilepsy2bids.annotations import Annotations


class TestAnnotations(unittest.TestCase):
    def test_loadTsv(self):
        annotations = Annotations.loadTsv("tests/sample.tsv")
        # Test loading of event
        self.assertEqual(len(annotations.events), 3)
        self.assertEqual(len(annotations.getEvents()), 3)
        self.assertEqual(len(annotations.getMask(1)), 3600)

    def test_saveTsv(self):
        annotations = Annotations.loadTsv("tests/sample.tsv")
        annotations.saveTsv("test.tsv")
        test = Annotations.loadTsv("test.tsv")
        self.assertEqual(len(test.events), len(annotations.events))
        np.testing.assert_array_equal(test.getMask(1), annotations.getMask(1))
        Path("test.tsv").unlink()

if __name__ == "__main__":
    unittest.main()


"""Script to download the SeizeIT dataset doi:10.48804/P5Q0OJ.
The script requires a valid API key and depends on the requests library.
"""

import hashlib
import json
import logging
import os
from pathlib import Path

import requests
from requests.adapters import HTTPAdapter, Retry


def downloadSeizeit(destination: str, apiToken: str) -> None:
    """Download the SeizeIT dataset.

    Args:
        destination (str): Path to the destination folder.
        apiToken (str): API key for the SeizeIT dataset.
    """

    SERVER_URL = "https://rdr.kuleuven.be"
    ID = "2407"
    VERSION = "1.0"

    logging.basicConfig(
        filename=os.path.join(destination, "downloadScript.log"),
        filemode="a",
        format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
        level=logging.DEBUG,
    )

    # Initialize session
    s = requests.Session()

    retries = Retry(total=5, backoff_factor=1)

    s.mount("http://", HTTPAdapter(max_retries=retries))
    s.mount("https://", HTTPAdapter(max_retries=retries))

    # Get list of Files
    url = SERVER_URL + "/api/datasets/" + ID + "/versions/" + VERSION + "/files"
    respFileList = s.get(url)
    if respFileList.ok:
        fileList = json.loads(respFileList.content)

        # Download each file
        for file in fileList["data"]:
            if "directoryLabel" not in file.keys():
                file["directoryLabel"] = "."
            # Create directory structure
            Path(os.path.join(destination, file["directoryLabel"])).mkdir(
                parents=True, exist_ok=True
            )
            filename = os.path.join(
                destination, file["directoryLabel"], file["dataFile"]["filename"]
            )

            # if file exists skip file:
            if os.path.exists(filename):
                md5Hash = hashlib.md5(open(filename, "rb").read()).hexdigest()
                if md5Hash == file["dataFile"]["md5"]:
                    logging.info("Skipping alreday dowloaded file {}".format(filename))
                    continue

            # Download file
            try:
                url = SERVER_URL + "/api/access/datafile/" + str(file["dataFile"]["id"])
                headers = {"X-Dataverse-key": apiToken}
                resp = s.get(url, headers=headers)
            except Exception as e:
                logging.error("{} - {}".format(file["dataFile"]["id"], e))
            if resp.ok:
                # Chec md5 hash
                md5Hash = hashlib.md5(resp.content).hexdigest()
                if md5Hash == file["dataFile"]["md5"]:
                    # Write file to disk
                    with open(filename, "wb") as edfFile:
                        edfFile.write(resp.content)
                else:
                    logging.error(
                        "{} - MD5 hash difference".format(file["dataFile"]["id"])
                    )
            else:
                logging.error(
                    "{} - Failed to download file : {}".format(
                        file["dataFile"]["id"], resp.status_code
                    )
                )
    else:
        logging.error(
            "Failed to retrieve dataset list status : {}".format(
                respFileList.status_code
            )
        )



"""load annotations from the CHBMIT dataset https://physionet.org/content/chbmit/1.0.0/ to a Annotations object."""

import os
import re

import pyedflib

from ..annotations import Annotation, Annotations, EventType, SeizureType


def _parseTimeStamp(string: str) -> float:
    """Parses timestamps from CHBMIT annotation files and returns a float representing the time from the earliest system time.

    Args:
        string (str): string to be parsed

    Returns:
        float: timestamp in seconds
    """
    timeStamp = re.findall(r"\d+", string)[-1]
    return float(timeStamp)


def _loadSeizures(edfFile: str, subject: str, edfFileName: str) -> list[tuple]:
    """Load seizures from a chb**-summary.txt file"""
    seizures = []
    summaryFile = os.path.join(
        os.path.dirname(edfFile), "{}-summary.txt".format(subject)
    )
    with open(summaryFile, "r") as summary:
        # Search for mention of edfFile in summary
        line = summary.readline()
        while line:
            if line == "File Name: {}\n".format(edfFileName):
                line = summary.readline()
                while ("File Name" not in line) and line:
                    if re.match(
                        "Seizure.*Start Time:", line
                    ):  # find start of new seizure
                        seizureStart = _parseTimeStamp(line)
                        seizureEnd = _parseTimeStamp(summary.readline())
                        seizures.append((seizureStart, seizureEnd))
                    line = summary.readline()
            else:
                line = summary.readline()
    return seizures


def loadAnnotationsFromEdf(edfFile: str) -> Annotations:
    """Loads annotations related to an EDF recording in the CHBMIT dataset.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Returns:
        Annotations: an Annotations object
    """
    # Subject - Select folder name as subject name
    subject = os.path.basename(os.path.dirname(edfFile))

    # dateTime and duration
    with pyedflib.EdfReader(edfFile) as edf:
        dateTime = edf.getStartdatetime()
        duration = edf.getFileDuration()
        edf._close()

    # Get Seizure type
    seizureType = SeizureType.sz  # seizure types are not available for CHB-MIT

    # Load Seizures
    seizures = _loadSeizures(edfFile, subject, os.path.basename(edfFile))

    # Confidence
    confidence = "n/a"  # confidence is not available for CHB-MIT

    # Channels
    channels = "n/a"  # channels are not available for CHB-MIT

    # Populate dictionary
    if len(seizures) == 0:
        seizureType = EventType.bckg
        seizures.append((0, duration))

    annotations = Annotations()
    for seizure in seizures:
        annotation = Annotation()
        annotation["onset"] = seizure[0]
        annotation["duration"] = seizure[1] - seizure[0]
        annotation["eventType"] = seizureType
        annotation["confidence"] = confidence
        annotation["channels"] = channels
        annotation["dateTime"] = dateTime
        annotation["recordingDuration"] = duration
        annotations.events.append(annotation)

    return annotations



"""Script to build a small sample EDF from an EDF recording. The recording header and list of signals are kept the same.
The content of data is white noise. The output file is named input_sample.edf
"""

import argparse
from pathlib import Path

import numpy as np
from pyedflib import highlevel

DURATION = 2
AMPLITUDE = 100


def makeSample(input: Path, outFilename: Path, duration: int = DURATION) -> None:
    signals, signal_headers, header = highlevel.read_edf(str(input))

    if isinstance(signals, np.ndarray):
        signals = signals.tolist()  # Make sure signals is a list

    # Keep only short duration of white noise
    for i in range(len(signals)):
        signals[i] = (
            np.random.rand(int(duration * signal_headers[i]["sample_frequency"])) * AMPLITUDE
        )

    highlevel.write_edf(str(outFilename), signals, signal_headers, header)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Shrink an EDF to a sample two seconds long",
        description="Shrinks an EDF to a sample. The recording header and list of signals are kept the same. "
        "The content of data is white noise. The output file is named input_sample.edf",
    )
    parser.add_argument("input", help="input edf file.")

    args = parser.parse_args()
    makeSample(args.input, args.input[:-4] + "_sample.edf". DURATION)



from importlib import resources as impresources
import json
import re
from shutil import rmtree
import subprocess
import unittest

from termcolor import cprint

from epilepsy2bids.bids.chbmit.convert2bids import convert as convertChbmit
from epilepsy2bids.bids.seizeit.convert2bids import convert as convertSeizeit
from epilepsy2bids.bids.siena.convert2bids import convert as convertSiena
from epilepsy2bids.bids.tuh.convert2bids import convert as convertTuh

TEST_DIR = impresources.files("tests") / "data"


class TestConvert(unittest.TestCase):
    def test_convert(self):
        for dataset, convert in zip(
            ("chbmit", "seizeit", "siena", "tuh"),
            (convertChbmit, convertSeizeit, convertSiena, convertTuh),
        ):
            convert(TEST_DIR / dataset, TEST_DIR / "bids" / dataset)
            rmtree(TEST_DIR / "bids" / dataset)
            cprint(
                f"Successfully converted {dataset.upper()} to BIDS.", "green", attrs=["bold"]
            )

    def test_bids_validator(self):
        for dataset, convert in zip(
            ("chbmit", "seizeit", "siena", "tuh"),
            (convertChbmit, convertSeizeit, convertSiena, convertTuh),
        ):
            datadir = TEST_DIR / "bids" / dataset
            convert(TEST_DIR / dataset, datadir)
            output = subprocess.run(
                [
                    "docker", "run", "-ti", "--rm", "-v", f"{datadir}:/data:ro", "bids/validator", "/data", "--ignoreSubjectConsistency", "--json"
                ],
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
            )
            if output.stdout != b"":
                try:
                    out = output.stdout.decode()
                    out = re.findall('{.*}', out)[0]
                    out = json.loads(out)
                    out = out["issues"]
                    if len(out["errors"]):
                        raise RuntimeError(
                            f"docker bids-validator returned the following errors : {out['errors']}"
                        )
                    if len(out["warnings"]):
                        raise RuntimeWarning(
                            f"docker bids-validator returned the following warnings : {out['warnings']}"
                        )
                except AttributeError:
                    continue
            rmtree(TEST_DIR / "bids" / dataset)

            cprint(
                f"Successfully run BIDS-validator on {dataset.upper()}.",
                "green",
                attrs=["bold"],
            )



import os
from importlib import resources as impresources
from pathlib import Path

from ... import bids
from ...bids.convert2bids import BidsConverter
from ...load_annotations.tuh import loadAnnotationsFromEdf

BIDS_DIR = impresources.files(bids)
DATASET = BIDS_DIR / "tuh"


def convert(root: Path, outDir: Path):
    root = Path(root)
    outDir = Path(outDir)
    bidsConverter = BidsConverter(
        BIDS_DIR, DATASET, root, outDir, loadAnnotationsFromEdf
    )

    subjectIdPairs = {}

    for subset in ["train", "dev", "eval"]:
        subRoot = root / subset
        for subjectFolder in sorted(subRoot.glob("*")):
            # Extract subject & session ID

            # Build dictionary of ID pairs
            if subjectFolder.name in subjectIdPairs.keys():
                subject = subjectIdPairs[subjectFolder.name]["subject"]
            else:
                subject = f"{len(subjectIdPairs):03}"
                subjectIdPairs[subjectFolder.name] = {}
                subjectIdPairs[subjectFolder.name]["subject"] = subject
                subjectIdPairs[subjectFolder.name]["subset"] = subset
                subjectIdPairs[subjectFolder.name]["session"] = {}

            for sessionFolder in sorted(subjectFolder.glob("*")):
                if sessionFolder.name in subjectIdPairs[subjectFolder.name]["session"]:
                    session = subjectIdPairs[subjectFolder.name]["session"][
                        sessionFolder.name
                    ]
                else:
                    session = f"{len(subjectIdPairs[subjectFolder.name]['session']):02}"
                    subjectIdPairs[subjectFolder.name]["session"][
                        sessionFolder.name
                    ] = session

                edfFiles = sorted((root / sessionFolder).glob("**/*.edf"))
                bidsConverter.buildBIDSHierarchy(edfFiles, subject, session)

    # Build participant metadata
    participants = {"participant_id": [], "TUH_id": [], "split": []}
    for folder in sorted(outDir.glob("sub-*")):
        subject = os.path.split(folder)[-1]
        participants["participant_id"].append(subject)
        subject = subject[4:]
        tuhId = list(subjectIdPairs.keys())[
            [x["subject"] for x in subjectIdPairs.values()].index(subject)
        ]
        participants["TUH_id"].append(tuhId)
        participants["split"].append(subjectIdPairs[tuhId]["subset"])

    bidsConverter.saveMetadata(participants)



"""load annotations from the Siena dataset https://physionet.org/content/siena-scalp-eeg/1.0.0/ to a Annotations object."""

import os
import re
import time
from pathlib import Path

import pandas as pd
import pyedflib

from ..annotations import Annotation, Annotations, EventType, SeizureType
from ..eeg import Eeg


def _parseTimeStamp(string: str) -> float:
    """Parses timestamps from Siena annotation files and returns a float representing the time from the earliest system time.

    Accepts the following formats with any random string before the timestamp :
    - 'Registration start time: 19.00.44'
    - 'Start time: 21:51:02'

    Args:
        string (str): string to be parsed as a timestamp

    Returns:
        float: timestamp in seconds from the earliest system time
    """
    timeStamp = re.findall(".*[ |:]([0-9]+.[0-9]+.[0-9]+).*", string)[0]
    return time.mktime(time.strptime(timeStamp.replace(":", "."), "%H.%M.%S"))


def _substractTimeStamps(time1: float, time2: float) -> float:
    """Substract time2 timestamp from time1 in seconds"""
    difference = time1 - time2
    if difference < 0:
        difference += 24 * 60 * 60  # Correct for day shift
    return difference


def _loadSeizures(edfFile: str) -> list[tuple]:
    """Load seizures from a Siena Seizures-list-Pxx.txt file

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Raises:
        ValueError: raises a ValueError if txt file format is unknown.

    Returns:
        seizures: A list of (start, end) tuples for each seizure in seconds from the beginning of the file.
    """

    subject = os.path.basename(os.path.dirname(edfFile))
    correctedEdfFileName = _correctEdfFileNameTypos(edfFile)

    seizures = []
    summaryFile = Path(os.path.dirname(edfFile)) / f"Seizures-list-{subject}.txt"
    with open(summaryFile, "r") as summary:
        # Search for mention of edfFile in summary
        line = summary.readline()
        while line:
            if re.search(r"File name: {} *\n".format(correctedEdfFileName), line):
                firstLine = summary.readline()
                # PN01 exception
                if correctedEdfFileName == "PN01.edf":
                    registrationStart = _parseTimeStamp(firstLine)
                    _ = summary.readline()
                    _ = summary.readline()
                    for _ in range(2):
                        _ = summary.readline()
                        start = _parseTimeStamp(summary.readline())
                        end = _parseTimeStamp(summary.readline())
                        _ = summary.readline()
                        seizures.append(
                            (
                                _substractTimeStamps(start, registrationStart),
                                (_substractTimeStamps(end, registrationStart)),
                            )
                        )
                elif correctedEdfFileName == "PN00-3.edf":  # seizure last 1hour
                    registrationStart = _parseTimeStamp(firstLine)
                    _ = summary.readline()
                    start = _parseTimeStamp(summary.readline())
                    end = _parseTimeStamp(summary.readline())
                    end = _parseTimeStamp(
                        "Seizure end time: 18.29.29\n"
                    )  # correct time
                    seizures.append(
                        (
                            _substractTimeStamps(start, registrationStart),
                            (_substractTimeStamps(end, registrationStart)),
                        )
                    )
                elif correctedEdfFileName == "PN10-7.8.9.edf":  # seizure last 1hour
                    if (
                        firstLine == "Registration start time:1 6.49.25\n"
                    ):  # extra space
                        firstLine = "Registration start time:16.49.25\n"
                    registrationStart = _parseTimeStamp(firstLine)
                    _ = summary.readline()
                    start = _parseTimeStamp(summary.readline())
                    end = _parseTimeStamp(summary.readline())
                    seizures.append(
                        (
                            _substractTimeStamps(start, registrationStart),
                            (_substractTimeStamps(end, registrationStart)),
                        )
                    )
                # PN12 exception
                elif "Seizure start time: " in firstLine:
                    start = _parseTimeStamp(firstLine)
                    end = _parseTimeStamp(summary.readline())
                    seizures.append(
                        (
                            _substractTimeStamps(start, registrationStart),
                            (_substractTimeStamps(end, registrationStart)),
                        )
                    )
                # Standard format
                elif "Registration start time:" in firstLine:
                    registrationStart = _parseTimeStamp(firstLine)
                    _ = summary.readline()
                    start = _parseTimeStamp(summary.readline())
                    end = _parseTimeStamp(summary.readline())
                    seizures.append(
                        (
                            _substractTimeStamps(start, registrationStart),
                            (_substractTimeStamps(end, registrationStart)),
                        )
                    )
                else:
                    raise ValueError("Unknown format for Siena summary file.")
            line = summary.readline()
    return seizures


def _getSeizureType(edfFile: str) -> SeizureType:
    """Get seizure type associated to an EDF file from the subject_info.csv file.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Raises:
        ValueError: raises a ValueError if the seizure type is unknown.

    Returns:
        SeizureType: the seizure type of the subject (Siena only records one seizure type per subject).
    """
    # Load CSV
    subject_info = pd.read_csv(Path(edfFile).parents[1] / "subject_info.csv")
    subject = os.path.basename(os.path.dirname(edfFile))

    seizureType = subject_info[subject_info.patient_id == subject][" seizure"].iloc[0]
    if seizureType == "IAS":
        seizureType = SeizureType.sz_foc_ia
    elif seizureType == "WIAS":
        seizureType = SeizureType.sz_foc_a
    elif seizureType == "FBTC":
        seizureType = SeizureType.sz_foc_f2b
    else:
        raise ValueError(f"Unknown seizure type ({seizureType}) for Siena dataset.")

    return seizureType


def _getChannels(edfFile: str) -> list[str]:
    """Get channels associated with a seizure from the subject_info.csv file.

    The subject_info.csv file provides lateralization and localization information for each subject. Electrodes from the
    10-20 system associated with the lateralization and localization are returned.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Raises:
        ValueError: raises a ValueError if the lateralization or localization is unknown.

    Returns:
        channels: List of electrodes names from the 10-20 system associated associated with a seizure.
    """
    # Load CSV
    subject_info = pd.read_csv(Path(edfFile).parents[1] / "subject_info.csv")
    subject = os.path.basename(os.path.dirname(edfFile))

    localization = subject_info[subject_info.patient_id == subject][
        " localization"
    ].iloc[0]
    lateralization = subject_info[subject_info.patient_id == subject][
        " lateralization"
    ].iloc[0]
    channels = Eeg.ELECTRODES_10_20
    if localization == "T":  # Temporal
        r = re.compile("^T.*")
    elif localization == "F":  # Frontal
        r = re.compile("^F.*")
    else:
        raise ValueError(f"Unknown localization for Siena dataset: {localization}.")
    channels = list(filter(r.match, channels))

    if lateralization == "R":
        r = re.compile(".*[02468]$")
    elif lateralization == "L":
        r = re.compile(".*[13579]$")
    elif lateralization == "Bilateral":
        r = re.compile(".*")
    else:
        raise ValueError(f"Unknown lateralization for Siena dataset: {lateralization}.")
    channels = list(filter(r.match, channels))

    return channels


def _correctEdfFileNameTypos(filename: str) -> str:
    """Correct typos in filenames in .txt file when compared to .edf files"""
    correctedEdfFileName = os.path.basename(filename)
    if correctedEdfFileName == "PN01-1.edf":
        correctedEdfFileName = "PN01.edf"
    elif correctedEdfFileName == "PN06-1.edf":
        correctedEdfFileName = "PNO6-1.edf"  # big o instead of 0
    elif correctedEdfFileName == "PN06-2.edf":
        correctedEdfFileName = "PNO6-2.edf"  # big o instead of 0
    elif correctedEdfFileName == "PN06-4.edf":
        correctedEdfFileName = "PNO6-4.edf"  # big o instead of 0
    elif correctedEdfFileName == "PN11-1.edf":
        correctedEdfFileName = "PN11-.edf"
    return correctedEdfFileName


def loadAnnotationsFromEdf(edfFile: str) -> Annotations:
    """Loads annotations related to an EDF recording in the Siena dataset.

    Args:
        edfFile (str): full path to the EDF for which annotations should be extracted.

    Returns:
        Annotations: an Annotations object
    """
    # dateTime and duration
    with pyedflib.EdfReader(edfFile) as edf:
        dateTime = edf.getStartdatetime()
        duration = edf.getFileDuration()
        edf._close()

    # Get Seizure type
    seizureType = _getSeizureType(edfFile)

    # Load Seizures
    seizures = _loadSeizures(edfFile)

    # Confidence
    confidence = "n/a"

    # Channels
    channels = _getChannels(edfFile)

    # Populate dictionary
    if len(seizures) == 0:
        seizureType = EventType.bckg
        seizures.append((0, duration))

    annotations = Annotations()
    for seizure in seizures:
        annotation = Annotation()
        annotation["onset"] = seizure[0]
        annotation["duration"] = seizure[1] - seizure[0]
        annotation["eventType"] = seizureType
        annotation["confidence"] = confidence
        annotation["channels"] = channels
        annotation["dateTime"] = dateTime
        annotation["recordingDuration"] = duration
        annotations.events.append(annotation)

    return annotations
